# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# pyrgsspecplot.py

from .version import VERSION, SAS_RELEASE, SAS_AKA

__version__ = f'pyrgsspecplot (pyrgsspecplot-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]' 

# pyrgsspecplot


import numpy as np
import pysas.pyutils.pyutils as pyutils
import pysas.pysasplot_utils.pysasplot_utils as sasplt
import sys
from astropy.io import fits
from astropy.table import Table
import time
from pysas.logger import TaskLogger as TL
from astropy import units as u
import matplotlib.pyplot as plt
from astropy import constants as const
from astropy.coordinates import Angle
import mpl_toolkits.axisartist as AA
from mpl_toolkits.axes_grid1 import host_subplot


logger = TL('pyrgsspecplot')


##################################
######### PLOTTING UTILS #########
##################################

def plot_spectrum(spec_data, axisunits, fits_info, t_labels, output, plot_info):
    """
    Plots the given spectrum passed as a table object.

    Args:
        spec_data: the table containing the data to be plotted.
        axisunits: the units to plot for the x-axis.
        fits_info: a list containing information to add to the plot.
        t_labels: the labels from the table, related to spec_data.
        output: the name of the plot file.

    Output:
        1 when finished.
    """
    
    fig, ax1 = plt.subplots()
    t_xlabel, t_ylabel, t_elabel = t_labels

    spec_data[t_elabel][np.isnan(spec_data[t_elabel])] = 0
#    ax1.errorbar(x = spec_data[t_xlabel].tolist(), y = spec_data[t_ylabel].tolist(), yerr = spec_data['STAT_ERR'].tolist(), barsabove = True, ecolor = 'black', linewidth = 0.3)
    ax1.set_ylabel('COUNTS/S')
    ax1.plot(spec_data[t_xlabel], spec_data[t_ylabel], label = 'Counts/second')

    ax1.plot(spec_data[t_xlabel], spec_data[t_elabel], label = 'Error')
    ax1.legend()

    obs_id, exp_id, telescope, instr, date_obs, date_end, object_fits, ra, dec, exp_idstr, order, exposuret, src_label = fits_info

    text = 'XMM. {0}. OBJECT: {1} RA: {2}. DEC: {3}\n'.format(instr, object_fits, ra, dec)
    text = text + 'OBS-ID: {0}. EXP-ID: {1}. Exp. Time: {2}.\nSource label: {3}'.format(obs_id, exp_id, exp_idstr, src_label)
    text = text + ''
    
    rebin_tag, alpha0, d, offaxis = plot_info

    fig.text(0.5, 0.8, text)
    ax1.set_position([0.1, 0.1, 0.8, 0.6])
    text_l = 'DATE-OBS: {0}. \nDATE-END: {1}.\n{2}.'.format(date_obs, date_end, rebin_tag)
    fig.text(0.1, 0.8, text_l)
    
    ax1.set_xlim(min(spec_data[t_xlabel]), max(spec_data[t_xlabel]))
    if axisunits == 'rad':
        ax1.set_xlabel('BETA CHANNEL')
        #ax2 = ax1.twiny()
        #ax2.set_position([0.1, 0.1, 0.8, 0.6])
        #ax2.spines['top'].set_position(('axes', 1))
        #ticks = ax1.get_xticks()
        #ticks[np.where(ticks == 0)] = 1
        #lambda_val = beta2lambda(d/10, alpha0, ticks, order * (-1), offaxis[0])
        
        #ax2.set_xticks(lambda_val)
        
        #ax2.set_xlim(min(lambda_val), max(lambda_val))

        #ax3 = ax1.twiny()        
        #ax3.set_position([0.1, 0.1, 0.8, 0.6])
        #ax3.spines['top'].set_position(('axes', 1.1))
        #energy_val = form_energy_axis(lambda_val)
        #ax3.set_xticks(energy_val)
    elif axisunits == 'angstrom':
        ax1.set_ylabel(r"$\lambda / \AA$")
        ax2 = ax1.twiny()
        ax2.set_position([0.1, 0.1, 0.8, 0.6])
    
    fig.set_size_inches(12, 9)

    fig.suptitle('rgsspecplot (Python version -- draft')
    if pyutils.is_notebook():
        plt.show()
    plt.savefig(output + '1')


def get_more_recent_ccf(abspaths):
    """
    Returns the absolute path of the most recent CCF file.

    Args:
        abspaths: the absolute path of the files to evaluate.

    Output:
        the absolute path of the most recent file.
    """

    maxn = 0
    for f in abspaths:
        f_aux = os.path.basename(f)
        num = int(re.search(r'\d+', f_aux).group(0))
        if num > maxn:
            maxn = num
            maxf = f

    return maxf


def check_xmm_misc(spec_header):
    """
    Checks the XMM miscellanous data CCF file to search for RGS values.

    Args:
        spec_header: the header of the spectrum.
        instr: the name of the instrument.

    Output:
        (d, alpha0): the line separation and the main angle.
    """

    logger.log('debug', 'Looking for platescale...')
    
    instr = spec_header['INSTRUME']

    d = 0
    alpha0 = 0

    try:
        alpha0 = spec_header['RGSALPHA']
    except KeyError:
        alpha0 = 0
    try:
        d = spec_header['LINE_SEP']
    except KeyError:
        d = 0

    if d != 0 and alpha0 != 0:
        return d, alpha0

    try:
        ccfpath = os.environ['SAS_CCFPATH']
    except KeyError:
        logger.log('error', 'Could not locate variable SAS_CCFPATH. Quitting...')
        sys.exit(0)

    folders = ccfpath.split(':')

    if len(folders) == 1:
        xmm_ccf_file = glob.glob(os.environ['SAS_CCFPATH'] + '/XMM_MISCDATA*')
        xmm_ccf_file = get_more_recent_ccf(xmm_ccf_file)
        logger.log('debug', 'CCF file used: {0}.'.format(xmm_ccf_file))
        with fits.open(xmm_ccf_file) as xmm:
            try:
                alpha0_data = xmm[1].data[xmm[1].data['PARM_ID'] == 'INCIDENCE_ANGLE'] 
                alpha0 = alpha0_data[alpha0_data['INSTRUMENT_ID'] == instr.upper()]['PARM_VAL'][0]
                alpha0_unit = alpha0_data[alpha0_data['INSTRUMENT_ID'] == instr.upper()]['PARM_UNIT'][0]
                if alpha0_unit.lower() == 'deg':
                    alpha0 = alpha0 * np.pi / 180
                d_data = xmm[1].data[xmm[1].data['PARM_ID'] == 'GRAT_LINE_DENS']
                d = d_data[d_data['INSTRUMENT_ID'] == instr.upper()]['PARM_VAL'][0]
            except (KeyError, IndexError) as e:
                logger.log('warning', 'Could not locate alpha0 and d in the XMM CCF file.')
    else:
        xmm_ccf_list = []
        logger.log('debug', 'List of folders found in CCF path: {0}'.format(folders))
        for folder in folders:
            try:
                xmm_ccf_file = os.path.abspath(glob.glob(folder + '/XMM_MISCDATA*')[0])
            except IndexError:
                logger.log('debug', '{} does not contain XMM misc. data calibration file.'.format(folder))
                continue
            xmm_ccf_list.append(xmm_ccf_file)
        logger.log('debug', 'Found the following XMM misc. data CCF compatible files: {0}'.format(xmm_ccf_list))
        latest_ccf = get_more_recent_ccf(xmm_ccf_list)
        logger.log('debug', 'CCF file used: {0}.'.format(latest_ccf))

        with fits.open(latest_ccf) as xmm:
            try:
                alpha0_data = xmm[1].data[xmm[1].data['PARM_ID'] == 'INCIDENCE_ANGLE']
                alpha0 = alpha0_data[alpha0_data['INSTRUMENT_ID'] == instr.upper()]['PARM_VAL'][0]
                alpha0_unit = alpha0_data[alpha0_data['INSTRUMENT_ID'] == instr.upper()]['PARM_UNIT'][0]
                if alpha0_unit.lower() == 'deg':
                    alpha0 = alpha0 * np.pi / 180
                d_data = xmm[1].data[xmm[1].data['PARM_ID'] == 'GRAT_LINE_DENS']
                d = d_data[d_data['INSTRUMENT_ID'] == instr.upper()]['PARM_VAL'][0]
            except (KeyError, IndexError) as e:
                logger.log('warning', 'Could not locate d and alpha0 in the XMM CCF file.')

    if alpha0 == 0 or d == 0:
        logger.log('error', 'Could not locate d or alpha0 neither in the Spectrum file or the XMM CCF file.')
        sys.exit(0)

    logger.log('info', 'Line separation and alpha angle retrieved from ccf file. Values: {}, {}.'.format(alpha0, d))
    
    return(float(d), float(alpha0))
    

def beta2lambda(d, alpha0, beta_values, order, offaxis):
    """
    Return the lambda equivalence to the beta channel array in the input.

    Args:
        d: the line separation of the spectrometer.
        alpha0: the angle of the spectrometer.
        lambda_values: the array containing the channel values.
        order: the order of the spectrum.
        offaxis: the value from the axis.

    Output:
        lamb_values: the lambda equivalence to the beta channel input.
    """
    
    if order == 0:
        logger.log('warning', 'Cannot obtain lambda values with order 0')
        return np.zeros(len(beta_values))

    lamb_values = (np.cos(beta_values) - np.cos(alpha0 + offaxis)) * (d / order)
    lamb_values = const.h.value * const.c.value / (beta_values * 1.602e-19) * 1e10

    print('values used:', lamb_values, alpha0, offaxis, d, order)
    return lamb_values


def form_energy_axis(lambda_val):
    """
    Convert an array of wavelengths into an array of energy using keV.

    Args:
        lambda_val: an array containing the wavelengths.

    Output:
        energy: the energy array.
    """
    
    lambda_val = lambda_val * u.angstrom
    energy = const.h * const.c / (lambda_val.to(u.meter))
    energy = energy.to(u.eV) / 1000

    return energy.value


def lambda2beta(d, alpha0, lambda_values, order, offaxis):
    """
    Transforms the beta channel values into wavelength measurements.

    Args:
        d: the line separation of the spectrometer.
        alpha0: the angle of the spectrometer.
        lambda_values: the array containing the channel values.
        order: the order of the spectrum.
        offaxis: the value from the axis.

    Output:
        beta: the beta conversion.
    """

    cosbeta = np.cos(alpha0 + offaxis) + order * lambda_values / d
    for i in range(0, len(cosbeta)):
        if cosbeta[i] > 1:
            cosbeta[i] = 0

    beta = np.arccos(cosbeta)
    print(d, alpha0, beta, order, offaxis)

    return beta


def run(iparsdic):
    print(f'Executing {__file__} {iparsdic}')
    t_start = time.time()

    group = iparsdic['group']
    if 'T' in group.upper() or 'Y' in group.upper():
        group = True
    else:
        group = False

    srclist = iparsdic['sourcelist']
    sourceid = int(iparsdic['sourceid'])
    plotfile = iparsdic['plotfile']
    rebin = iparsdic['rebin']

    spectrumsets = iparsdic['spectrumsets']
    if ',' in spectrumsets:
        spectrumsets = spectrumsets.replace(' ', '')
        spectrumsets = spectrumsets.split(',')
    else:
        spectrumsets = spectrumsets.split(' ')

    nspectrumsets = len(spectrumsets)
    if nspectrumsets == 0 or nspectrumsets > 2:
        logger.log('error', 'Invalid number of spectrums. Either 1 or 2.')
        sys.exit(0)

    if 'T' in rebin.upper() or 'Y' in rebin.upper():
        rebin = True
    else:
        rebin = False

    n_points = int(iparsdic['n_points'])

    for spectrum in spectrumsets:
        nspec = 0
        
        if nspectrumsets == 1:
            pass
        else:
            nspec = nspec + 1
            plotfile = plotfile + '_{0}'.format(nspec)
            
        logger.log('info', 'Evaluating file {}.'.format(spectrum))
        # collect fits info TODO
        with fits.open(spectrum) as f:
            spec_data = f[1].data
            logger.log('debug', 'Filtering data using pyutils...')
            spec_data = Table(spec_data)
            spec_data = pyutils.filter_data(spec_data)
            spec_header = f[1].header

        try:
            hduclas2 = spec_header['HDUCLAS2']
        except KeyError:
            logger.log('warning', 'Keyword HDUCLAS2 missing in the spectrum file {}. Selecting BACKGROUND as default.'.format(spectrum))
            spectrumtype = 'BACKGROUND'
        if hduclas2 == 'TOTAL':
            spectrumtype = 'SRC+BKG'
        elif hduclas2 == 'NET':
            spectrumtype = 'NET'
        else:
            spectrumtype = 'BACKGROUND'

        try:
            tcuni = spec_header['TCUNI']
        except KeyError:
            try:
                tcuni = spec_header['TCUNI1']
            except KeyError:
                logger.log('warning', 'Keyword TCUNI missing in the spectrum file {}. Selecting rad as default.'.format(spectrum))
            tcuni = 'rad'
            axisunits = 'BACKGROUND'
        if tcuni == 'Angstrom':
            axisunits = 'angstrom'
        elif tcuni == 'rad':
            axisunits = 'rad'
        else:
            axisunits = 'rad'

        src_label, offaxis = get_src_label_and_offaxes(srclist, sourceid)
        exposuret = float(pyutils.get_key_word(spectrum, 'EXPOSURE'))
        order = spec_header['RFLORDER']

        try:
            spec_delta = spec_header['SPECDELT']
        except KeyError:
            spec_delta = 1
        try:
            specpix = spec_header['SPECPIX']
        except KeyError:
            specpix = 0
        try:
            specval = spec_header['SPECVAL']
        except KeyError:
            specval = 0

        detchans = spec_header['DETCHANS']
        nbins = len(spec_data)

        try:
            betaref = spec_header['TCRVL']
        except KeyError:
            try:
                betaref = spec_header['TCRVL1']
            except KeyError:
                logger.log('error', 'Could not locate reference in the input file.')
                sys.exit(0)
        try:
            betawid = spec_header['TCDLT']
        except KeyError:
            try:
                betawid = spec_header['TCDLT1']
            except KeyError:
                logger.log('error', 'Could not locate width reference in the input file.')
                sys.exit(0)

        channel_p = spec_data['CHANNEL']

        try:
            counts_error_p = spec_data['STAT_ERR']
        except KeyError:
            pass 
        t_elabel = 'STAT_ERR'
        t_xlabel = 'CHANNEL'

        try:
            counts_p = spec_data['COUNTS']
            t_ylabel = 'COUNTS'

            try:
                counts_error_p = spec_data['STAT_ERR']
            except KeyError:
                spec_data['STAT_ERR'] = np.zeros(len(counts_p))
        except KeyError:
            try:
                rate_p = spec_data['RATE']
                spec_data['RATE'] = exposuret * rate_p
                try:
                    spec_data['STAT_ERR'] = abs(spec_data['STAT_ERR'])
                except KeyError:
                    spec_data['STAT_ERR'] = np.sqrt(abs(spec_data['RATE']))

                #rate_to_counts = rate_p * exposuret
                #ratetocountserr = counts_error_p * exposuret
                t_ylabel = 'RATE'
            except KeyError:
                logger.log('error', 'Could not detect either COUNTS or RATE columns in spectrum data.')
                sys.exit(0)

        try:
            counts_error_p = spec_data['STAT_ERR']
        except KeyError:
            if t_ylabel == 'RATE':
                pass
            elif t_ylabel == 'COUNTS':
                pass
        spec_data['STAT_ERR'] = np.sqrt(abs(spec_data['RATE']))
        
        if group:
            bins = 0
            summed_counts = 0
            summed_channels = 0
            n_channels = 0

        if rebin:
            DTin, DTout = sasplt.get_time_deltas(spectrum, 1, n_points, 'CHANNEL')
            rebin_tag = 'Rebin: Delta relation: {0}/{1}'.format(DTin, DTout)
            tstable = sasplt.TSrebin(spec_data, DTin, DTout, t_xlabel, t_ylabel, t_elabel)
        else:
            DTin, DTout = 1, 1
            rebin_tag = 'No rebinning'
            tstable = sasplt.TSrebin(spec_data, DTin, DTout, t_xlabel, t_ylabel, t_elabel)

        fits_info = collect_fits_info(spectrum)

        d, alpha0 = check_xmm_misc(spec_header)
        plot_info = (rebin_tag, alpha0, d, offaxis)
        fits_info.append(order)
        fits_info.append(exposuret)
        fits_info.append(src_label)
        t_labels = (t_xlabel, t_ylabel, t_elabel)
        
        plot_spectrum(tstable, axisunits, fits_info, t_labels, plotfile, plot_info)

    t_stop = time.time()

    logger.log('info', 'Finished running pyrgsspecplot in {} seconds.'.format(round(t_stop - t_start, 2)))


def get_src_label_and_offaxes(srclist, sourceid):
    """
    Returns the label and the offaxis values matching the source id from the source list.

    Args:
        srclist: the FITS file containing the source list info.
        sourceid: the source id that is being evaluated.

    Output:
        src_label: the label taken from the source file.
        offaxis: tuple with the DELTA_DISP and the DELTA_XDSP values.
    """

    logger.log('debug', 'Getting source label and offaxis values from {} with id {}.'.format(srclist, sourceid))

    with fits.open(srclist) as f:
        try:
            src_label = f['SRCLIST'].data['LABEL'][sourceid - 1]
        except (KeyError, IndexError) as e:
            logger.log('warning', 'Could not locate label with the current source id ({0}) and source list ({1})'.format(sourceid, srclist))
            src_label = 'Indef'
        try:
            src_delta_disp = f['SRCLIST'].data['DELTA_DISP'][sourceid - 1]
            src_delta_xdsp = f['SRCLIST'].data['DELTA_XDSP'][sourceid - 1]
            src_delta_disp = Angle(src_delta_disp, u.arcminute)
            src_delta_xdsp = Angle(src_delta_xdsp, u.arcminute)

            offaxis = (src_delta_disp.radian, src_delta_xdsp.radian)
        except (KeyError, IndexError) as e:
            logger.log('warning', 'Could not locate the offaxis values within the current source list and source id.')
            offaxis = (0,0)

    return src_label, offaxis


def collect_fits_info(specfits):
    """
    Returns the needed information from a spectrum fits file.

    Args:
        specfits: the spectrum fits file, either the path or an already opened
    astropy object.

    Output:
        obs_id, exp_id, telescope, instr, date_obs, date_end, object_fits, ra, dec,
    exp_idstr
    """

    obs_id = pyutils.get_key_word(specfits, 'OBS_ID')
    exp_id = pyutils.get_key_word(specfits, 'EXP_ID')
    telescope = pyutils.get_key_word(specfits, 'TELESCOP')
    instr = pyutils.get_key_word(specfits, 'INSTRUME')
    date_obs = pyutils.get_key_word(specfits, 'DATE-OBS')
    date_end = pyutils.get_key_word(specfits, 'DATE-END')
    object_fits = pyutils.get_key_word(specfits, 'OBJECT')

    if object_fits == 'unknown':
        object_fits = 'Indef'

    ra = pyutils.get_key_word(specfits, 'RA_OBJ')
    if ra == 'unknown':
        ra = pyutils.get_key_word(specfits, 'RA')
        if ra == 'unknown':
            ra = 'Indef'

    dec = pyutils.get_key_word(specfits, 'DEC_OBJ')
    if dec == 'unknown':
        dec = pyutils.get_key_word(specfits, 'DEC')
        if dec == 'unknown':
            dec = 'Indef'

    exp_idstr = pyutils.get_key_word(specfits, 'EXPIDSTR')

    return[obs_id, exp_id, telescope, instr, date_obs, date_end, object_fits, ra, dec, exp_idstr]

